from __future__ import annotations

from datetime import datetime, timedelta, timezone
from enum import Enum
import logging

from sqlbind import AND_, WHERE, WITH, EMPTY, Expr, QExpr, QueryParams, not_empty
from typing import Any, Callable, cast

from eportal.db import session
from eportal.db.models import DAY_DELTA
from eportal.dbutils import QueryList, concat

log = logging.getLogger(__name__)


class BaseFilter:
    class Operation(Enum):
        IN = 'in list'
        NOT_IN = 'not in list'
        LIKE = 'contains'
        EQUALS = 'equals'
        SMALLER = 'smaller than'
        GREATER = 'greater than'

    joins: list[str] = []
    field: str
    name: str

    def __init__(self, oper: 'BaseFilter.Operation') -> None:
        self.operation = oper

        self.oper_map: dict['BaseFilter.Operation', Callable[..., Expr]] = {
            self.Operation.IN: QueryParams.IN,
            self.Operation.NOT_IN: lambda *args: ~(QueryParams.IN(*args)),
            self.Operation.LIKE: QueryParams.LIKE,
            self.Operation.EQUALS: QueryParams.eq,
            self.Operation.SMALLER: lambda q, v, e: QExpr(q, v).__lt__(e),
            self.Operation.GREATER: lambda q, v, e: QExpr(q, v).__gt__(e),
        }

        self.params_map: dict['BaseFilter.Operation', Callable[[str], tuple]] = {
            self.Operation.IN: self._array_params,
            self.Operation.NOT_IN: self._array_params,
            self.Operation.LIKE: self._like_params,
            self.Operation.EQUALS: self._single_param,
            self.Operation.SMALLER: self._single_param,
            self.Operation.GREATER: self._single_param,
        }

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and self.operation == other.operation

    def apply(self, query: QueryParams, value: str) -> Expr:
        return self.oper_map[self.operation](query, self.field, *self.params_map[self.operation](value))

    @staticmethod
    def value_adapter(value: str) -> Any:
        return value  # pragma: no cover

    def _like_params(self, value: str) -> tuple[str, Expr]:
        if value.startswith('='):
            return '{}', not_empty / value[1:].lower()
        if value.startswith('^'):
            return '{}%', not_empty / value[1:].lower()
        return '%{}%', not_empty / value.lower()

    def _array_params(self, value: str) -> tuple[list[Expr]]:
        return (not_empty / [self.value_adapter(it) for it in value.split(',')],)

    def _single_param(self, value: str) -> tuple[Expr]:
        return (not_empty / self.value_adapter(value),)


class IntBaseFilter(BaseFilter):
    @staticmethod
    def value_adapter(value: str) -> Any:
        return int(value)


class TagInFilter(IntBaseFilter):
    name = 'Tag'

    def __init__(self, inverse: bool = False) -> None:
        super(TagInFilter, self).__init__(BaseFilter.Operation.NOT_IN if inverse else BaseFilter.Operation.IN)
        self.inverse = inverse

    def apply(self, query: QueryParams, value: str) -> Expr:
        if not value:
            return EMPTY
        in_expr = query.IN('stag.tag_id', *self._array_params(value))
        expr = Expr(f'EXISTS (SELECT 1 FROM kc_servers_tag stag WHERE stag.server_id = s.server_id {AND_(in_expr)})')
        return ~expr if self.inverse else expr


class TagValueLikeFilter(BaseFilter):
    name = 'Tag value'

    def __init__(self) -> None:
        super(TagValueLikeFilter, self).__init__(BaseFilter.Operation.LIKE)

    def apply(self, query: QueryParams, value: str) -> Expr:
        if not value:
            return EMPTY
        expr = query.LIKE('lower(stag.tag_value)', *self._like_params(value))
        return Expr(f'EXISTS (SELECT 1 FROM kc_servers_tag stag WHERE stag.server_id = s.server_id {AND_(expr)})')


class BUnitFilter(IntBaseFilter):
    name = 'Business Unit'
    joins = ['LEFT JOIN kc_keys k ON s.kc_key = k.key']
    field = 'k.unit_id'


class DaysInactiveFilter(IntBaseFilter):
    name = 'Days inactive'
    field = '(julianday(current_timestamp) - julianday(coalesce(s.last, 0)))'


class ServersStats:
    def __init__(self, filters: list[tuple[BaseFilter, str]] | None = None) -> None:
        self.query = session.query_params()
        joins: list[str] = []
        where: list[Expr] = []
        for flt, value in filters or []:
            joins += [j for j in flt.joins if j not in joins]
            where.append(flt.apply(self.query, value))

        self.servers_cte = f's AS (SELECT s.* FROM kc_servers s {concat(*joins)} {WHERE(*where)})'

    def _execute(self, sql: str, as_dict: bool = False) -> QueryList[tuple] | QueryList[dict]:
        return self.query.execute(f'{WITH(self.servers_cte)} {sql}', as_dict=as_dict)

    def get_servers_count(self) -> int:
        return cast(int, self._execute('SELECT count(server_id) FROM s').scalar())

    def get_uptime_stats(self) -> tuple[int, dict[str, int | None]]:
        """
        Returns pair of (total_servers_number, {uptime_aggregate: uptime_seconds})
        Where uptime_aggregate = (max|average|median|m0|m6|m12|m18)
        """

        ranges = [int(timedelta(days=30 * months).total_seconds()) for months in (6, 12, 18)]
        m0_count_expr = f'count(CASE WHEN {self.query.uptime < ranges[0]} THEN 1 ELSE NULL END)'
        m6_count_expr = f'count(CASE WHEN {self.query.in_range("uptime", ranges[0], ranges[1])} THEN 1 ELSE NULL END)'
        m12_count_expr = f'count(CASE WHEN {self.query.in_range("uptime", ranges[1], ranges[2])} THEN 1 ELSE NULL END)'
        m18_count_expr = f'count(CASE WHEN {self.query.uptime >= ranges[2]} THEN 1 ELSE NULL END)'
        sql = f'''SELECT max(uptime) max, cast(avg(uptime) as integer) average, count(server_id) total_servers,
            {m0_count_expr} m0, {m6_count_expr} m6, {m12_count_expr} m12, {m18_count_expr} m18 FROM s'''
        data = self._execute(sql, as_dict=True)[0]
        count = cast(int, data.pop('total_servers'))
        if count:
            limit = self.query.LIMIT(2 - count % 2)
            offset = self.query.OFFSET((count - 1) // 2)
            sql = f'SELECT cast(avg(uptime) as integer) m FROM (SELECT uptime FROM s ORDER BY uptime {limit} {offset})'
            data['median'] = self._execute(sql).scalar()
        else:
            data['median'] = None
        return count, data

    def get_distro_stats(self) -> list[tuple[str, int, int]]:
        """Returns records [distro, kcare_active_servers, kcare_silent_servers]"""

        active_time = datetime.now(tz=timezone.utc) - timedelta(days=DAY_DELTA)
        distro_expr = "coalesce(coalesce(nullif(distro, '') || '-' || nullif(distro_version, ''), distro), 'unknown')"
        active_count_expr = f'count(CASE WHEN {self.query.last > active_time} THEN 1 ELSE NULL END)'
        silent_count_expr = f'count(CASE WHEN last is NULL OR {self.query.last <= active_time} THEN 1 ELSE NULL END)'
        sql = f'''SELECT {distro_expr} distro, {active_count_expr} active_cnt, {silent_count_expr} silent_cnt
                  FROM s JOIN kc_keys k ON k.key = s.kc_key
                  WHERE k.products IS NULL OR 'kernel' IN (SELECT value FROM json_each(products))
                  GROUP BY distro, distro_version
                  ORDER BY active_cnt DESC, silent_cnt DESC, distro'''
        return self._execute(sql)

    def get_arch_stats(self) -> list[tuple[str, int]]:
        """Returns records [arch, number_of_servers]"""

        sql = """SELECT coalesce(nullif(machine, ''), 'unknown') arch, count(*) cnt
                 FROM s GROUP BY machine ORDER BY cnt DESC, arch"""
        return self._execute(sql)
