from __future__ import annotations

import hashlib
import re
import logging
from datetime import datetime
from typing import Any

from eportal import utils, db, dbutils, patches_constants as pc
from eportal.delivery_kit import CheckinPackage
from eportal.db.models import registered_server_ids, get_all_kcare_checkins_as_dict, get_all_libcare_checkins_as_dict
from eportal.patches.api import get_release_by_kernel_level, get_all_releases_as_dict

SQLITE_DATETIME_FMT = '%Y-%m-%d %H:%M:%S.%f'
logger = logging.getLogger(__name__)


def strip_version(name: str) -> str | None:
    if name:
        return re.split(r'([_-]\d|\.so)', name)[0]
    return None  # pragma: no cover TODO


@utils.lru_cache_with_expire(10000, 300)
def get_kernel_release(kernel_id: str, level: int) -> str | None:
    return get_release_by_kernel_level(pc.Stream.KERNEL, kernel_id, level)


@utils.lru_cache_with_expire(10000, 300)
def get_userspace_release(stream_type: str, build_id: str, level: int) -> str | None:
    if stream := pc.Stream.items.get(stream_type):
        return get_release_by_kernel_level(stream, build_id, level)
    return None  # pragma: no cover, return for typechecker


def save_kernel_stat(data: list[dict[str, Any]]) -> None:
    from statsd.defaults.env import statsd

    cleaned: list[dict] = list(filter(None, (prepare_kernel_checkin(it) for it in data)))
    if not cleaned:
        return

    flist = list(cleaned[0].keys())
    flist.remove('server_id')

    # NOTE: this is parametrized query without possible SQL injection
    set_expr = ', '.join(f'{f}=@{f}' for f in flist)
    q = f'UPDATE kc_servers SET {set_expr} WHERE server_id = @server_id'

    with dbutils.transaction(db.session):
        with dbutils.cursor(db.session) as cur:
            for checkin in cleaned:
                cur.execute(q, checkin)

    statsd.incr('eportal.success_checkins_kernel', len(cleaned))


def prepare_kernel_checkin(data: dict[str, Any]) -> dict[str, Any] | None:
    server_id = data.get('server_id')
    if not server_id:
        return None

    return dict(
        server_id=server_id,
        node=data['node'],
        distro_version=data['distro_version'],
        distro=data['distro'],
        machine=data['machine'],
        processor=data['processor'],
        release=data['release'],
        version=data['version'],
        kcare_version=data['kcare_version'],
        virt=data['virt'],
        euname=data['euname'],
        remote_ip=data['ip'],
        patch_level=data['patch_level'],
        kernel_id=data['kernel_id'],
        uptime=data['uptime'],
        patch_type=data['patch_type'],
        ltimestamp=data['ltimestamp'] and datetime.utcfromtimestamp(data['ltimestamp']).strftime(SQLITE_DATETIME_FMT),
        last=datetime.utcfromtimestamp(data['last']).strftime(SQLITE_DATETIME_FMT),
        patchset=get_kernel_release(data['kernel_id'], data['patch_level']),
    )


def save_userspace_stat(data: list[dict[str, Any]]) -> None:
    from statsd.defaults.env import statsd

    existing_server_ids = registered_server_ids([it['server_id'] for it in data])
    cleaned = [prepare_userspace_checkin(it) for it in data if it['server_id'] in existing_server_ids]
    skipped_count = len(data) - len(cleaned)
    if skipped_count:
        logger.debug('skipped %d userspace stat entries after server_id check', skipped_count)

    if not cleaned:
        return

    flist = list(cleaned[0].keys())

    columns_expr = ', '.join(flist)
    values_expr = ', '.join(f'@{f}' for f in flist)
    sql = f'INSERT OR REPLACE INTO kc_ucheckins ({columns_expr}) VALUES ({values_expr})'

    with dbutils.transaction(db.session):
        with dbutils.cursor(db.session) as cur:
            for checkin in cleaned:
                cur.execute(sql, checkin)

    statsd.incr('eportal.success_checkins_userspace', len(cleaned))


def prepare_userspace_checkin(data: dict[str, Any]) -> dict[str, Any]:
    server_id = data.get('server_id')
    return dict(
        server_id=server_id,
        product=pc.Stream.items[data['stream']].product.value,
        libname=strip_version(data['lib']),
        build_id=data['build_id'],
        patch_level=data['patch_level'],
        kcare_version=data['kcare_version'],
        last=datetime.utcfromtimestamp(data['last']).strftime(SQLITE_DATETIME_FMT),
        patchset=get_userspace_release(data['stream'], data['build_id'], data['patch_level']),
    )


def depersonalize(data: list[dict[str, Any]], sensitive_keys: tuple[str, ...]) -> list[dict[str, Any]]:
    for entry in data:
        for key in sensitive_keys:
            if key in entry:
                # NOTE: usedforsecurity=False option is available since Python 3.9
                entry[key] = hashlib.md5(str(entry[key]).encode()).hexdigest()  # nosec B324 B303
    return data


def make_checkin_info_package(since: datetime | None = None, until: datetime | None = None) -> CheckinPackage:
    with CheckinPackage() as pkg:
        pkg.add_json('releases.json', {'data': get_all_releases_as_dict()}, compact=True)

        sensitive_keys = ('server_id', 'remote_ip', 'node')
        checkins = depersonalize(get_all_kcare_checkins_as_dict(since, until), sensitive_keys=sensitive_keys)
        pkg.add_json('kcare_checkins.json', {'data': checkins}, compact=True)

        ucheckins = depersonalize(get_all_libcare_checkins_as_dict(since, until), sensitive_keys=sensitive_keys)
        pkg.add_json('libcare_checkins.json', {'data': ucheckins}, compact=True)
        return pkg
