from __future__ import annotations

import os
import shutil
import subprocess
import tempfile
from functools import wraps
from itertools import chain, repeat
from pathlib import Path
from runpy import run_path
from typing import Any, Callable, Literal, overload

import config
import distro
import requests
from requests.auth import HTTPBasicAuth

from eportal.backups import backup_databases
from eportal.delivery_kit import DoctorPackage
from eportal.patches_constants import Stream
from eportal.patchsource import PatchSource
from eportal.utils import get_version, get_proxy_settings

EXTERNAL_URLS = [
    'https://patches.kernelcare.com',
    'https://repo.cloudlinux.com/kernelcare/kernelcare_install.sh',
    'https://repo.eportal.kernelcare.com/',
    'https://downloads.kernelcare.com/',
]
GET_MY_IP_URL = 'https://patches.kernelcare.com/myip'

EPORTAL_USER_CONFIG = '/etc/eportal/config'


def handle_exceptions(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except Exception as error:
            return {'error': repr(error)}

    return wrapper


def compact_patches_info(feeds: list[str], dashboard: dict[str, Any]) -> dict[str, Any]:
    status = dashboard['status']
    unpacked = dashboard['unpacked']
    patchsets = dashboard['patchsets']
    stitle = {None: 'Undeployed', True: 'Enabled', False: 'Disabled', 'NA': 'N/A'}

    ps_state: dict[str, dict[str, list[str]]] = {}
    for ps in patchsets:
        for f in feeds:
            key = f, ps
            if key in status:
                s = stitle.get(status[key], 'UNKNOWN')
            else:
                s = 'Undeployed' if ps in unpacked else 'Not downloaded'
            ps_state.setdefault(ps, {}).setdefault(s, []).append(f)

    return ps_state


@handle_exceptions
def get_patches_info() -> dict[str, dict[str, Any]]:
    # NOTE: eportal.web importing forces DB client initialization.
    # In case importing globally and DB files are corrupted it will fail to load.

    from eportal.patches_impl import get_repo
    from eportal.web.views.main import get_dashboard_content

    patches_info = {}
    dc = get_dashboard_content(0, None)
    for stream in Stream:
        repo = get_repo(0, stream)
        stat = repo.dashboard(dc.feeds, usage=False)
        patches_info[stream.value] = compact_patches_info(dc.feeds, stat)
    return patches_info


def get_resource(url: str, auth: HTTPBasicAuth | None = None) -> tuple[int, dict, bytes]:
    s = requests.Session()
    s.proxies = get_proxy_settings()
    r = s.get(url, auth=auth, timeout=10)
    return r.status_code, dict(r.headers), r.content


def check_proxy() -> bool:
    for envvar in list(os.environ):
        if envvar.lower() in ('http_proxy', 'https_proxy'):
            return True
    return False


def check_connectivity() -> dict[str, dict[str, Any] | bool]:
    result: dict[str, dict[str, Any] | bool] = {'proxy': check_proxy()}

    ps = PatchSource()
    upload_url = [(ps.get_url(DoctorPackage.upload_uri), HTTPBasicAuth(*ps.auth))] if ps.is_configured() else []
    for url, auth in chain(zip(EXTERNAL_URLS, repeat(None)), upload_url):
        try:
            status, headers, _ = get_resource(url, auth=auth)
            result[url] = {'status': status, 'headers': headers}
        except Exception as error:
            result[url] = {'error': repr(error)}
    return result


@overload
def run_command(command: str, text_output: Literal[True] = True) -> tuple[str, str]: ...  # pragma: no branch


@overload
def run_command(command: str, text_output: Literal[False]) -> tuple[bytes, bytes]: ...  # pragma: no branch


def run_command(command: str, text_output: bool = True) -> tuple[str, str] | tuple[bytes, bytes]:
    result = subprocess.run(command, shell=True, capture_output=True)  # nosec: B602
    if not text_output:
        return result.stdout, result.stderr
    return result.stdout.decode('utf8', errors='replace'), result.stderr.decode('utf8', errors='replace')


@handle_exceptions
def get_disk_usage() -> list[str]:
    output, _ = run_command('df -h')
    return output.split('\n')


@handle_exceptions
def get_os_version() -> dict[str, str]:
    distro_id = distro.id()
    distro_version = distro.version()
    distro_name = distro.name()
    return {'ID': distro_id, 'Version': distro_version, 'Name': distro_name}


@handle_exceptions
def get_process_list() -> list[str]:
    result, _ = run_command('ps aux | grep -E "(nginx|eportal)"')
    return result.split('\n')


def collect_logs(pkg: DoctorPackage, subdir: str) -> None:
    for nginx_log in ('access.log', 'kcare-eportal.log', 'error.log'):
        stdout, stderr = run_command(f'tail -n 10000 /var/log/nginx/{nginx_log}', text_output=False)
        pkg.add_file(f'{subdir}/{nginx_log}', data=stdout + stderr)
    stdout, stderr = run_command('journalctl -u eportal -ocat -n 10000', text_output=False)
    pkg.add_file(f'{subdir}/eportal.log', data=stdout + stderr)
    stdout, stderr = run_command('tail -n 10000 /var/log/audit/audit.log', text_output=False)
    pkg.add_file(f'{subdir}/audit.log', data=stdout + stderr)


def is_sensitive_data_key(key: str) -> bool:
    return any(pattern in key for pattern in ('KEY', 'SECRET', 'DATABASE', 'PROXY', 'PASSWORD'))


def prepare_config(path: str) -> dict:
    config = run_path(path)
    hidden_value = '********'
    return {k: hidden_value if is_sensitive_data_key(k) else v for k, v in config.items() if k.isupper()}


@handle_exceptions
def get_config_options() -> dict[str, Any]:
    paths = [config.jroot('config', 'local.py'), config.jroot('config', 'docker.py'), EPORTAL_USER_CONFIG]
    eportal_config: dict[str, Any] = {}
    for path in paths:
        if os.path.exists(path):
            conf = prepare_config(path)
        else:
            conf = {'msg': 'No user defined config options.'}
        eportal_config[path] = conf
    return eportal_config


@handle_exceptions
def get_eportal_ip() -> str:
    _, _, content = get_resource(GET_MY_IP_URL)
    return content.decode('utf-8').strip()


def get_eportal_database(pkg: DoctorPackage, subdir: str) -> None:
    tmpdir = tempfile.mkdtemp()
    try:
        backup_databases(None, tmpdir)
        pkg.add_file(subdir, filepath=Path(tmpdir))
    except Exception as error:
        pkg.add_file(f'{subdir}/error.log', data=b'Failed to copy DB: ' + repr(error).encode())
    finally:
        shutil.rmtree(tmpdir, ignore_errors=True)


def make_doctor_report() -> DoctorPackage:
    version, platform = get_version()
    report = {
        'eportal_ip': get_eportal_ip(),
        'os_version': get_os_version(),
        'eportal_version': version,
        'platform': platform,
        'patches_info': get_patches_info(),
        'connectivity': check_connectivity(),
        'disk_usage': get_disk_usage(),
        'process_list': get_process_list(),
        'config': get_config_options(),
    }
    with DoctorPackage() as pkg:
        pkg.add_json('report.json', report)
        collect_logs(pkg, 'logs')
        get_eportal_database(pkg, 'db')
        return pkg
