from __future__ import annotations

import json
import time
from datetime import datetime
import logging
from typing import Any, Union, TypeVar

from sqlalchemy.sql.elements import quoted_name

from eportal import checkin_saver, oplog, audit, patches_constants as pc, dynconfig
from eportal.utils import TaskOpts, adapt_args, iter_chunks
from eportal.db import models as m, session as main_session
from eportal.dbutils import transaction
from eportal.patches import api


UNKNOWN_USER = 'UNKNOWN'

logger = logging.getLogger(__name__)

ReplicatedModel = TypeVar(
    'ReplicatedModel',
    bound=Union[m.User, m.Feed, m.KcKey, m.KcServer, m.Unit, m.UnitUser, m.Tag, m.ApiKey],
)

TABLE_MODELS = {M.__tablename__: M for M in (m.User, m.Feed, m.KcKey, m.KcServer)}


# deprecated, need for compatibility
@oplog.replicate('server-checkin', oplog.CLEAN_POLICY_AGGRESSIVE)  # pragma: no cover
def checkin_server(data: dict[str, Any]) -> None:
    checkin_saver.save_kernel_stat([data])


@oplog.replicate('server-checkin-many', oplog.CLEAN_POLICY_AGGRESSIVE)
def checkin_server_many(data: list[dict[str, Any]]) -> None:
    checkin_saver.save_kernel_stat(data)


@oplog.replicate('userspace-checkin', oplog.CLEAN_POLICY_AGGRESSIVE)
def checkin_userspace(data: list[dict[str, Any]]) -> None:
    checkin_saver.save_userspace_stat(data)


@oplog.register_action('server-register')
def register_server_by_id(server_id: str, ip: str, hostname: str, key: str) -> None:
    m.add_server(ip, hostname, key, server_id=server_id)


def register_server(ip: str, hostname: str, key: str, replicate: bool = True) -> m.KcServer:
    server = m.add_server(ip, hostname, key)
    params = {
        'server_id': server.server_id,
        'ip': ip,
        'hostname': hostname,
        'key': key,
    }
    replicate and register_server_by_id.write(**params)
    return server


@oplog.replicate('server-unregister')
def unregister_server(server_id: str) -> None:
    return m.delete_server(server_id)


@oplog.replicate('server-unregister-by-key-and-ip')
def unregister_server_by_key_and_ip(key: str, ip: str) -> int:
    return m.delete_server_by_key_and_ip(key, ip)


class LegacyMltag:
    def __init__(self, mltag: str, unit_id: int = 0) -> None:
        self.tags = self._parse_mltag(mltag)
        self.unit_id = unit_id

    def _parse_mltag(self, mltag: str) -> dict[str, str | None]:
        def parse_name_value_pair(item: str) -> tuple[str, str | None] | None:
            name, *value = item.split(':', 1)
            name = name.strip()
            return (name, value[0].strip() if value else None) if name else None

        return dict(filter(None, (parse_name_value_pair(item) for item in mltag.split(';'))))

    def to_tag_list(self, create_non_existent: bool = True, user: str = UNKNOWN_USER) -> list[tuple[int, str | None]]:
        current_tags = {t.name: t.tag_id for t in main_session.query(m.Tag).filter_by(unit_id=self.unit_id).all()}
        tag_list = [(current_tags[name], value) for name, value in self.tags.items() if name in current_tags]
        if create_non_existent:  # pragma: no branch
            with transaction(main_session):
                for name in self.tags.keys() - current_tags.keys():
                    new_tag = m.Tag(name=name, unit_id=self.unit_id)
                    main_session.add(new_tag)
                    main_session.flush()
                    sync_model(new_tag, user=user)
                    tag_list.append((new_tag.tag_id, self.tags[name]))

        return tag_list


@oplog.replicate('server-tag')
@adapt_args(rename={'tags': 'mltag'}, defaults={'unit_id': 0, 'user': UNKNOWN_USER})
def tag_server(
    server_id: str,
    unit_id: int,
    user: str,
    mltag: str | None = None,
    tag_list: list[tuple[int, str | None]] | None = None,
) -> bool:
    # TODO: do we need audit log instead?
    logger.info('Tagging server %s with "%s", user: %s, unit_id: %d', server_id, tag_list or mltag, user, unit_id)

    if mltag is not None:
        if not unit_id:
            kc_server = main_session.get(m.KcServer, server_id)
            unit_id = kc_server.kc_key_rel.unit_id if kc_server and kc_server.kc_key_rel else 0
        tag_list = LegacyMltag(mltag, unit_id=unit_id).to_tag_list(user=user)

    return m.tag_server(server_id, tag_list=tag_list or [], unit_id=unit_id)


@oplog.replicate('tag-servers')
def tag_servers(tag_id: int, servers: list[str], value: str | None = None) -> int:
    return m.tag_servers(tag_id, servers, value)


@oplog.replicate('untag-servers')
def untag_servers(tag_id: int, servers: list[str]) -> int:
    return m.untag_servers(tag_id, servers)


@oplog.replicate('manage-sync-table-record')
@transaction(main_session)
def sync_table_record(table: str | quoted_name, fields: dict[str, Any]) -> None:
    q = main_session.query_params()
    q.execute(f'REPLACE INTO {table} {q.VALUES(**fields)}')


@oplog.replicate('manage-delete-table-record-v2')
@transaction(main_session)
def delete_table_record(model_name: str, pk: Any) -> None:
    obj = main_session.get(getattr(m, model_name), pk)
    if obj:
        m.delete_object(obj)


@oplog.replicate('manage-delete-table-record')
@transaction(main_session)
def delete_table_record_legacy(table: str, pk: str, value: Any, fix_keys_for_feed: str | None = None) -> None:
    _ = pk
    pkvalue: Any
    if table == 'feed':
        pkvalue = {'unit_id': 0, 'name': value}
    else:
        pkvalue = value
    delete_table_record.action(TABLE_MODELS[table].__name__, pkvalue)


def sync_model(model: ReplicatedModel, user: str = UNKNOWN_USER) -> None:
    data = model.to_json(public=False)
    sync_table_record.action.write(table=model.__tablename__, fields=data)

    if hasattr(model, 'audit_change_msg'):  # pragma: no branch
        # only KcServer has no audit_change_msg method, but it can't be passed to sync_model
        audit_log(getattr(model, 'unit_id', 0), user, *model.audit_change_msg())


@transaction(main_session)
def delete_model(model: ReplicatedModel, user: str = UNKNOWN_USER) -> None:
    m.delete_object(model)

    pk = {col.name: getattr(model, col.name) for col in model.__table__.primary_key}
    delete_table_record.action.write(model_name=model.__class__.__name__, pk=pk)

    if hasattr(model, 'audit_delete_msg'):
        audit_log(getattr(model, 'unit_id', 0), user, *model.audit_delete_msg())


@oplog.replicate('manage-deploy-patchset')
@adapt_args(defaults={'unit_id': 0}, rename={'store': 'stream'})
def deploy_patchset(
    stream: str,
    unit_id: int,
    patchset: str,
    deploy_orders: dict[str, str],
    user: str | None = None,
) -> None:
    from eportal.patches_impl import get_repo
    from eportal import tasks

    repo = get_repo(unit_id, pc.Stream.items[stream])
    for feed, status in deploy_orders.items():
        if status == 'disable':
            repo.rollback_patches(feed, patchset, user=user)
        elif status == 'enable':
            tasks.deployment_worker_single.push(
                TaskOpts(urgent=True), unit_id=unit_id, feed=feed, patchset=patchset, store_type=stream, user=user
            )
        elif status == 'enable-upto':
            tasks.deployment_worker.push(
                TaskOpts(urgent=True), unit_id=unit_id, feed=feed, patchset=patchset, store_type=stream, user=user
            )
        elif status == 'undeploy-downto':
            repo.undeploy_down_to(feed, patchset, user=user)


def get_registrations() -> dict[str, list[str] | list[list[str]]]:
    fields = ['kc_key', 'server_id', 'remote_ip', 'node']
    sql = f'SELECT {", ".join(fields)} FROM kc_servers'
    tag_fields = ['server_id', 'tag_id', 'tag_value']
    tag_sql = f'SELECT {", ".join(tag_fields)} FROM kc_servers_tag'
    return {
        'fields': fields,
        'result': main_session.query_params().execute(sql),
        'tag_fields': tag_fields,
        'tag_result': main_session.query_params().execute(tag_sql),
    }


def _bulk_insert(table: str, fields: list[str], data: list, bulk_size: int = 1000) -> None:
    # session.add() is too slow to use directly, here we use sqlite json
    # functions to upsert records with 50x speedup

    q = main_session.query_params()
    jfields = ', '.join(f"json_extract(jd.value, '$[{i}]')" for i in range(len(fields)))
    sql = f'''\
        REPLACE INTO {table} ({', '.join(fields)})
        SELECT {jfields}
        FROM json_each(:data) jd
    '''

    for chunk in iter_chunks(data, bulk_size):
        with transaction(main_session):
            q['data'] = json.dumps(chunk)  # type: ignore[call-arg]
            q.execute(sql)


def apply_server_registrations(registrations: dict[str, list]) -> None:
    _bulk_insert('kc_servers', registrations['fields'], registrations['result'])
    if 'tag_fields' in registrations:  # pragma: no branch
        _bulk_insert('kc_servers_tag', registrations['tag_fields'], registrations['tag_result'])


def full_changes() -> oplog.FullChanges:
    return {
        'users': [it.to_json(public=False) for it in main_session.query(m.User)],
        'feeds': [it.to_json(public=False) for it in main_session.query(m.Feed)],
        'keys': [it.to_json(public=False) for it in main_session.query(m.KcKey)],
        'units': [it.to_json(public=False) for it in main_session.query(m.Unit)],
        'unit_users': [it.to_json(public=False) for it in main_session.query(m.UnitUser)],
        'api_keys': [it.to_json(public=False) for it in main_session.query(m.ApiKey)],
        'tags': [it.to_json(public=False) for it in main_session.query(m.Tag)],
        'patches': {it.value: api.instance_release_deploy_status(it) for it in pc.REPLICATED_STREAMS},
        'config': dynconfig.Config().data,
    }


def apply_full_changes(changes: oplog.FullChanges, audit_user: str = 'AUTO') -> None:
    from eportal import tasks

    for it in changes.get('units', []):
        sync_table_record.action(m.Unit.__tablename__, it)

    for it in changes.get('unit_users', []):
        sync_table_record.action(m.UnitUser.__tablename__, it)

    for it in changes.get('users', []):
        sync_table_record.action(m.User.__tablename__, it)

    for it in changes.get('feeds', []):
        sync_table_record.action(m.Feed.__tablename__, it)

    for it in changes.get('keys', []):
        sync_table_record.action(m.KcKey.__tablename__, it)

    for it in changes.get('api_keys', []):
        sync_table_record.action(m.ApiKey.__tablename__, it)

    for it in changes.get('tags', []):
        sync_table_record.action(m.Tag.__tablename__, it)

    config_data = changes.get('config')
    if config_data:  # pragma: no branch
        dynconfig.save_config_data.action(config_data)

    patches = changes.get('patches', {})
    for stream_type, remote_status in patches.items():
        stream = pc.Stream.items[stream_type]
        local_status = {
            (unit_id, feed, release): status
            for unit_id, feed, release, status in api.instance_release_deploy_status(stream)
        }
        for unit_id, feed, patchset, enabled in remote_status:
            local_enabled = local_status.get((unit_id, feed, patchset))
            if local_enabled == enabled:
                continue

            if local_enabled is None:
                audit_msg = (
                    f'downloading patchset {patchset} with enabled={enabled} to feed {feed} during reconciliation'
                )
            elif enabled:
                audit_msg = f'enabling patchset {patchset} in feed {feed} during reconciliation'
            else:
                audit_msg = f'disabling patchset {patchset} in feed {feed} during reconciliation'

            audit.add_record(unit_id, audit_user, audit.ACTION_MANAGE_PATCHSET, audit_msg)
            tasks.deployment_worker_single.push(
                TaskOpts(urgent=True),
                unit_id=unit_id,
                feed=feed,
                patchset=patchset,
                store_type=stream_type,
                enabled=enabled,
            )

        remote_patchsets = set((unit_id, feed, patchset) for unit_id, feed, patchset, _ in remote_status)
        for unit_id, feed, patchset in set(local_status) - remote_patchsets:
            audit.add_record(
                unit_id,
                audit_user,
                audit.ACTION_MANAGE_PATCHSET,
                f'removed {patchset} patchset from feed {feed} during reconciliation',
            )
            api.remove_releases_from_feed(unit_id, stream, [patchset], feed)


@oplog.replicate('server-cleanup')
def delete_servers(
    server_ids: list[str] | None = None,
    ips: list[str] | None = None,
    hosts: list[str] | None = None,
    last_check_in_age: int | None = None,
    unit_id: int | None = None,
    user: str = UNKNOWN_USER,
    audit_params: dict[str, Any] | None = None,
) -> int:
    count = m.delete_servers(
        server_ids=server_ids, ips=ips, hosts=hosts, last_check_in_age=last_check_in_age, unit_id=unit_id
    )
    if count:
        params = audit_params or {
            'server_ids': server_ids,
            'ips': ips,
            'hosts': hosts,
            'last_check_in_age': last_check_in_age,
            'unit_id': unit_id,
        }
        params = {key: value for key, value in params.items() if value is not None}
        details = f'{count} server{" was" if count == 1 else "s were"} deleted with params: {params}'
        _audit_log(unit_id or 0, user, audit.ACTION_MANAGE_SERVER, details, ts=int(time.time()))
    return count


@oplog.register_action('audit-log')
def _audit_log(unit_id: int, user: str, action: str, details: str, ts: int) -> None:
    audit.add_record(unit_id, user, action, details, datetime.utcfromtimestamp(ts))


def audit_log(unit_id: int, user: str, action: str, details: str) -> None:
    ops = dict(unit_id=unit_id, user=user, action=action, details=details, ts=int(time.time()))
    _audit_log(**ops)
    _audit_log.write(**ops)


@oplog.replicate('create-user')
@adapt_args(defaults={'unit_id': 0})
def get_or_create_user(
    unit_id: int, username: str, description: str, role: str = 'readonly'
) -> tuple[m.User | m.UnitUser, bool]:
    with transaction(main_session):
        user = m.get_user(username, unit_id)
        if user:
            return user, False

        if unit_id:
            user = m.UnitUser(  # nosec hardcoded_password_funcarg
                unit_id=unit_id, username=username, password_hash='', role=None, description=description
            )
        else:
            # empty password to prevent access after user deletion from identity provider
            user = m.User(  # nosec hardcoded_password_funcarg
                username=username, description=description, password_hash='', active=True
            )
            user.set_role(role)

        main_session.add(user)
        main_session.flush()

    return user, True
