from __future__ import annotations

import logging
import os
import time
import threading
from typing import cast, Any
from datetime import datetime, timezone

import config
from flask_login import current_user
from eportal import oplog, utils, replicated, patches_constants as pc
from eportal.checkin_saver import make_checkin_info_package
from eportal.utils import Scheduler, adapt_args, atomic_write
from eportal import signature
from eportal.patches_impl import get_repo
from eportal.spooler import task, task_runner, task_clean
from eportal.db import models

log = logging.getLogger('eportal.tasks')


@task('deployment-worker')
@adapt_args(defaults={'unit_id': 0})
def deployment_worker(feed: str, patchset: str, store_type: str, unit_id: int, user: str | None = None) -> None:
    log.info('Deployment started: feed < %s >, patchset < %s >' % (feed or 'default', patchset))
    get_repo(unit_id, pc.Stream.items[store_type]).deploy_up_to(feed, patchset, user=user)


@task('deploy-up-to')
@adapt_args(defaults={'unit_id': 0})
def deployment_worker_single(
    feed: str, patchset: str, store_type: str, unit_id: int, enabled: bool = True, user: str | None = None
) -> None:
    log.info('Deployment started (one): feed < %s >, patchset < %s >' % (feed or 'default', patchset))
    get_repo(unit_id, pc.Stream.items[store_type]).deploy_up_to(feed, patchset, upto=False, enabled=enabled, user=user)


@task('auto-update', context=True)
def auto_deployment_worker(task: dict[str, Any], store_type: str) -> None:
    log.info('Feeds auto update started...')
    unit_ids = [0] + models.Unit.unit_ids()
    for unit_id in unit_ids:
        get_repo(unit_id, pc.Stream.items[store_type]).update_feeds(deadline=cast(float, task['eta']), user='AUTO')


@task('send-report')
def reporting_worker() -> None:
    from eportal.report import send_report

    log.info('Reporting: started...')
    send_report()


@task('clean-inactive-servers')
def clean_inactive_servers_worker() -> None:
    from eportal.replicated import delete_servers

    log.info('Cleaning inactive servers: started...')
    deleted = delete_servers(last_check_in_age=30, user='AUTO')
    log.info(f'{deleted} server{" was" if deleted == 1 else "s were"} deleted')


@task('reconcile-patches')
def reconcile_patches() -> None:
    if oplog.leader_url_is_self(config.LEADER_URL):
        log.warning("Can't sync from self, try change/specify leader url")
        return

    changes = oplog.get_full_changes(config.LEADER_URL)
    changes = {'patches': changes['patches']}
    replicated.apply_full_changes(changes)


@task('update-signatures')
def update_signatures(user: str | None = None) -> None:
    user = user or (current_user.username if current_user else 'UNKNOWN')
    signature.update_signatures(user)


def _epoch_to_utc_datetime(epoch: float) -> datetime:
    return datetime.fromtimestamp(epoch, tz=timezone.utc).replace(tzinfo=None)


@task('send-checkin-report')
def send_checkin_report(state_file: str) -> None:
    since: datetime | None = None
    try:
        with open(state_file, 'r') as f:
            since = _epoch_to_utc_datetime(float(f.read()))
    except Exception as err:
        log.debug("Couldn't parse state file %s: %s", state_file, err)

    now = time.time()
    until = _epoch_to_utc_datetime(now)

    pkg = make_checkin_info_package(since=since, until=until)
    try:
        pkg_name = pkg.send()
        log.info('Checkin report sent successfully: %s', pkg_name)
        atomic_write(state_file, str(now))
    finally:
        pkg.remove()


def clean_oplog() -> None:
    from eportal import oplog

    now = time.time()
    oplog.clean(now - config.REPLICATION_CLEAN_AGE)
    oplog.clean(now - config.REPLICATION_CLEAN_AGGRESSIVE_AGE, oplog.CLEAN_POLICY_AGGRESSIVE, vacuum=True)


def run_scheduler() -> None:
    state_dir = os.path.join(config.BASE_DIR, 'config')
    utils.ensure_dir(state_dir)
    s = Scheduler(state_dir, log)
    if config.SERVER_STATS_PERIOD:
        s.add_task('send-report', config.SERVER_STATS_PERIOD, reporting_worker)
    else:
        log.info('Report sending is disabled.')

    if config.AUTO_UPDATE_PERIOD:
        s.add_task('auto-update', config.AUTO_UPDATE_PERIOD, auto_deployment_worker, store_type=pc.Stream.KERNEL.value)
        s.add_task(
            'auto-update-user', config.AUTO_UPDATE_PERIOD, auto_deployment_worker, store_type=pc.Stream.USER.value
        )
        s.add_task(
            'auto-update-qemu', config.AUTO_UPDATE_PERIOD, auto_deployment_worker, store_type=pc.Stream.QEMU.value
        )
        s.add_task('auto-update-db', config.AUTO_UPDATE_PERIOD, auto_deployment_worker, store_type=pc.Stream.DB.value)
        s.add_task('update-signatures', config.AUTO_UPDATE_PERIOD, update_signatures)

    if config.LEADER_URL:
        s.add_task('reconcile-patches', config.RECONCILE_PATCHES_PERIOD, reconcile_patches)

    if config.NODE_URL:
        s.add('clean-oplog', 3600, clean_oplog)

    if config.CLEAN_INACTIVE_SERVERS_PERIOD:
        s.add_task('auto-cleanup-inactive-servers', config.CLEAN_INACTIVE_SERVERS_PERIOD, clean_inactive_servers_worker)

    if config.SEND_CHECKIN_PERIOD and not config.REDUCED_REPORT:
        state_file = os.path.join(state_dir, 'send-checkin-report.stamp')
        send_period = max(4 * 3600, config.SEND_CHECKIN_PERIOD)  # 4 hours is check-in period for agents
        s.add_task('send-checkin-report', send_period, send_checkin_report, state_file)

    s.run()
    log.info('Scheduler is empty.')


def main() -> None:
    threading.Thread(name='scheduler', target=run_scheduler, daemon=True).start()

    oplog_sched = oplog.make_scheduler()
    threading.Thread(name='oplog-scheduler', target=oplog_sched.run, daemon=True).start()

    task_clean()
    task_runner()


if __name__ == '__main__':  # pragma: no cover
    main()
