from __future__ import annotations

import config
import copy
from datetime import datetime, timedelta
from typing import Any

import jwt

SIGN_ALGO = 'HS256'


def get_sign_secret() -> str:
    return config.REPLICATION_SHARED_KEY or str(config.SECRET_KEY)


class TokenError(Exception):
    pass


def create(data: dict[str, Any], audience: str, ttl: timedelta | None = None, now: datetime | None = None) -> str:
    data = copy.deepcopy(data)
    data['aud'] = [audience]
    data['iat'] = now or datetime.utcnow().replace(microsecond=0)
    if ttl is not None:
        data['exp'] = datetime.utcnow() + ttl

    return jwt.encode(
        payload=data,
        key=get_sign_secret(),
        algorithm=SIGN_ALGO,
    )


def verify(
    token: str, audience: str, allow_endless: bool = False, required_claims: list[str] | None = None
) -> dict[str, Any]:
    required_claims = list(required_claims) if required_claims else []

    if not allow_endless:
        required_claims.append('exp')

    try:
        data: dict[str, Any] = jwt.decode(
            token,
            key=get_sign_secret(),
            audience=audience,
            algorithms=[
                SIGN_ALGO,
            ],
        )

        # pyjwt handles None values as absent claims, so we have to implement our check allowing None values
        for claim in required_claims:
            if claim not in data:
                raise TokenError(f'Token is missing the "{claim}" claim')

        return data
    except jwt.PyJWTError as e:
        raise TokenError(e) from e
