from __future__ import annotations

import logging
import json
from datetime import datetime, timedelta, timezone
from functools import reduce
from urllib.parse import quote
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    cast,
    Generic,
    Literal,
    TypeVar,
    Union,
    TYPE_CHECKING,
)

from jinja2.runtime import Context
from flask import Response
from flask import g, url_for, redirect, flash, request
from flask_admin import Admin, AdminIndexView, expose
from flask_admin.actions import action
from flask_admin.babel import gettext, ngettext, lazy_gettext
from flask_admin.form import FormOpts
from flask_admin.form.rules import RuleSet
from flask_admin.contrib.sqla import ModelView
from flask_admin.form.fields import Select2Field
from flask_admin.contrib.sqla.filters import (
    FilterEqual,
    FilterInList,
    FilterLike,
    IntGreaterFilter,
    DateEqualFilter,
    DateGreaterFilter,
    DateSmallerFilter,
)
from flask_admin.contrib.sqla.tools import parse_like_term
from flask_admin.model.base import ViewArgs, OrderedDict
from flask_admin.model.helpers import get_mdict_item_or_list
from flask_admin.helpers import get_redirect_target
from flask_admin.model.template import TemplateLinkRowAction
from flask_login import current_user
from jinja2 import pass_context
from markupsafe import Markup
from sqlalchemy import and_, or_, text, func, case, distinct
from sqlalchemy.orm import Query, aliased, joinedload
from sqlalchemy.sql import exists
from wtforms.fields import SelectField, StringField, PasswordField
from wtforms.validators import NumberRange, DataRequired, NoneOf, Length, Optional as OptionalV, Regexp
from wtforms.form import Form
from markupsafe import escape

import config
from eportal import db, perms, replicated, audit, patches_constants as pc
from eportal.db.models import (
    DEFAULT_FEED,
    Base,
    User,
    KcKey,
    Feed,
    KcServer,
    KcServerTag,
    Tag,
    Unit,
    UnitUser,
    ApiKey,
    get_unique_id,
    get_tags,
)
from eportal.utils import iter_chunks
from eportal.web import app, login_manager, units
from eportal.audit import AuditLog
from .forms import ServerTagsForm

logger = logging.getLogger(__name__)
T = TypeVar('T')
UM = TypeVar('UM', bound=Union[KcKey, Feed, AuditLog, UnitUser, Tag])
RM = TypeVar('RM', bound=Union[User, KcKey, Feed, Unit, UnitUser, ApiKey, KcServer, Tag])

if not TYPE_CHECKING:  # pragma: no branch

    class ModelView(Generic[T], ModelView):
        pass


def formatter(fn: Callable) -> Callable:
    def fmt(_view: ModelView, _context: Context, model: Base | None, name: str) -> str:
        value = reduce(lambda obj, attr: getattr(obj, attr, None), name.split('.'), model)
        return fn(value)  # type: ignore[no-any-return]

    return fmt


def mformatter(fn: Callable) -> Callable:
    def fmt(_view: ModelView, _context: Context, model: Base, name: str) -> str:
        return fn(model)  # type: ignore[no-any-return]

    return fmt


@formatter
def key_name_formatter(value: str) -> Markup:
    # key name is a link to server list connected to this key
    return Markup("<a href='{0}'>{1}</a>".format(url_for('kcserver.index_view', flt_key_equals=value), escape(value)))


@mformatter
def server_limit_formatter(model: 'KcKey') -> str:
    # column shows the relation of registered servers to its key limit
    limit = model.server_limit
    return '{0} / {1}'.format(model.server_count(), limit if limit else 'unlimited')


@formatter
def feed_formatter(value: str) -> str:
    return value or DEFAULT_FEED


@mformatter
def products_formatter(model: 'KcKey') -> str:
    products_list = model.products_list or []
    products_labels = [pc.Product.titles.get(prod, prod) for prod in products_list]
    return ', '.join(products_labels)


@formatter
def channel_formatter(value: str) -> str:
    return pc.Channel.titles.get(value, value)


@formatter
def business_unit_formatter(name: str | None) -> str:
    return name or Markup('<i>Default</i>')


def filter_formatter(_view: "AuditView", _context: Context, model: AuditLog, name: str) -> Markup:
    value = getattr(model, name)
    url = f'.?flt_{name}_equals=' + quote(value)
    return Markup(f'<a href="{url}">{escape(value)}</a>')


def convert_utc_date_to_local(date: datetime) -> datetime:
    return date.replace(tzinfo=timezone.utc).astimezone(tz=None)


def convert_local_date_to_utc(date: datetime) -> datetime:
    return date.astimezone(tz=timezone.utc).replace(tzinfo=None)


@formatter
def local_date_formatter(value: datetime | None) -> str:
    if value:
        local_date = convert_utc_date_to_local(value)
        return local_date.strftime(config.DATETIME_FORMAT)
    return ''


class Choices:
    def __init__(self, get_names_func: Callable) -> None:
        self.get_names_func = get_names_func

    def __iter__(self) -> Iterable:
        values = list(self.get_names_func())
        if values and isinstance(values[0], tuple):
            return iter(values)

        choices = [(name, name) for name in values]
        return iter(choices)


PermArg = Union[perms.Perm, perms.PermSet]


def make_permissions(
    view: PermArg = perms.NONE, modify: PermArg = perms.NONE, **kwargs: PermArg
) -> dict[str, perms.PermSet]:
    result = {
        'index_view': perms.to_permset(view),
        'create_view': perms.to_permset(modify),
        'edit_view': perms.to_permset(modify),
        'delete_view': perms.to_permset(modify),
        'details_view': perms.to_permset(view),
        'action_view': perms.to_permset(modify),
        'export': perms.to_permset(view),
        'server_tags_view': perms.to_permset(modify),
    }
    result.update((k, perms.to_permset(v)) for k, v in kwargs.items())
    return result


def get_tag_filter_options(unit_id: int) -> list[tuple[int, str]]:
    is_super_admin = config.BUNITS and (unit_id == 0)
    _, tags = get_tags(None if is_super_admin else unit_id)
    if not tags:
        return [(0, '')]  # empty list or None will enable arbitrary input in select2

    def tag_option_title(tag: Tag) -> str:
        if is_super_admin:
            return f'{tag.name} ({tag.unit.title if tag.unit else "Default"})'
        return tag.name

    return [(tag.tag_id, tag_option_title(tag)) for tag in tags]


class BaseView(ModelView[T]):
    permissions = make_permissions()
    _export_columns: list[tuple[str, str]]

    def is_accessible(self) -> bool:
        return cast(bool, current_user.is_authenticated)

    def inaccessible_callback(self, name: str, **kwargs: Any) -> Response:
        if not current_user.is_authenticated:
            return cast(Response, login_manager.unauthorized())
        return super().inaccessible_callback(name, **kwargs)  # pragma: no cover TODO

    def _handle_view(self, name: str, **kwargs: Any) -> Response | None:
        if not self.has_access(name):
            return self.inaccessible_callback(name, **kwargs)
        return super()._handle_view(name, **kwargs)

    def has_access(self, name: str) -> bool:
        view_permissions = self.permissions.get(name, {perms.NONE})
        return perms.has_permissions(current_user.permissions, view_permissions)

    @property
    def can_create(self) -> bool:
        return self.has_access('create_view')

    @property
    def can_edit(self) -> bool:
        return self.has_access('edit_view')

    @property
    def can_delete(self) -> bool:
        return self.has_access('delete_view')

    def mutate_form(self, model: T | None, form: Form, is_created: bool) -> None:
        """Allows to modify form based on request data"""

    def mutate_form_opts(self, form_opts: FormOpts, is_created: bool) -> None:
        """Allows to modify form opts based on request data"""

    def edit_form(self, obj: T | None = None) -> Form:
        form = super().edit_form(obj=obj)
        self.mutate_form(obj, form, False)
        return form

    def create_form(self) -> Form:
        form = super().create_form()
        self.mutate_form(None, form, True)
        return form

    def list_columns_ex(self, columns: list[tuple[str, str]]) -> list[tuple[str, str]]:
        """Allows to define list columns based on request data

        columns MUST not be changed
        """
        return columns

    def details_columns_ex(self, model: T, columns: list[tuple[str, str]]) -> list[tuple[str, str]]:  # pragma: no cover
        """Allows to define details columns based on request data

        columns MUST not be changed
        """
        return columns

    def export_columns_ex(self, columns: list[tuple[str, str]]) -> list[tuple[str, str]]:  # pragma: no cover
        """Allows to define details columns based on request data

        columns MUST not be changed
        """
        return columns

    def render(self, template: str, **kwargs: Any) -> Response:
        ep = cast(str, request.endpoint)
        if ep.endswith('.index_view'):
            kwargs['list_columns'] = self.list_columns_ex(kwargs['list_columns'])
        elif ep.endswith('.details_view'):
            kwargs['details_columns'] = self.details_columns_ex(kwargs['model'], kwargs['details_columns'])
        elif ep.endswith('.edit_view'):
            self.mutate_form_opts(kwargs['form_opts'], is_created=False)
        elif ep.endswith('.create_view'):  # pragma: no branch
            self.mutate_form_opts(kwargs['form_opts'], is_created=True)
        return super().render(template, **kwargs)

    def _export_data(self) -> tuple[int, list[T]]:
        # flask-admin has no way to override export columns besides
        # rewriting self._export_columns. Here is a nasty trick
        # to save original _export_columns and reuse the value later.
        if not hasattr(self, '_eportal_export_columns'):
            self._eportal_export_columns = self._export_columns
        self._export_columns = self.export_columns_ex(self._eportal_export_columns)
        return super()._export_data()


class Select2FieldMultiple(Select2Field):
    def iter_choices(self) -> Iterator[tuple[str, str, bool]]:
        """fix check if list item is selected"""
        selected_items = self.data or []
        for value, label in self.choices:
            yield (value, label, value in selected_items)


class UnitBaseView(BaseView[UM]):
    unit_exclude_default = False
    form_overrides: dict[str, Any] | None
    form_args: dict[str, Any] | None
    form_edit_rules: Iterable[str] | None
    form_columns: Iterable[str] | None
    column_list: Iterable[str] | None
    column_sortable_list: Iterable[str] | None
    column_formatters: dict[str, Any] | None
    column_labels: dict[str, Any] | None

    def __init__(self, *args: Any, **kwargs: Any):
        form_overrides = self.form_overrides or {}
        form_overrides['unit_id'] = SelectField
        self.form_overrides = form_overrides

        form_args = self.form_args or {}
        form_args['unit_id'] = dict(choices=Choices(self.make_titles), label='Business Unit')
        self.form_args = form_args

        if self.form_edit_rules:
            self.form_edit_rules = ('unit_id',) + tuple(self.form_edit_rules)

        self.form_columns = ('unit_id',) + tuple(self.form_columns or ())

        self.column_formatters = dict(self.column_formatters or {}, **{'unit.title': business_unit_formatter})
        self.column_list = tuple(self.column_list or ()) + ('unit.title',)
        self.column_labels = dict(self.column_labels or {}, **{'unit.title': 'Business Unit'})
        self.column_sortable_list = tuple(self.column_sortable_list or ()) + ('unit.title',)

        super().__init__(*args, **kwargs)

    def make_titles(self) -> Iterable[tuple[int, str]]:
        titles = units.titles()
        if self.unit_exclude_default:
            titles = titles.copy()
            titles.pop(0)
        return titles.items()

    @classmethod
    def discard_bunit_column_when_required(cls, columns: list[tuple[str, str]]) -> list[tuple[str, str]]:
        if config.BUNITS and not units.current_unit():
            return columns
        # unit title should be the last column
        return columns[:-1]

    def list_columns_ex(self, columns: list[tuple[str, str]]) -> list[tuple[str, str]]:
        return self.discard_bunit_column_when_required(columns)

    def unit_filter_query(self, q: Query) -> Query:
        if (unit_id := units.current_unit_id()) or not config.BUNITS:
            q = q.filter_by(unit_id=unit_id)
        return q

    def get_one(self, id: str) -> UM | None:
        model = super().get_one(id)
        if unit_id := units.current_unit_id():
            if model and model.unit_id != unit_id:
                return None
        return model

    def get_query(self) -> Query:
        return self.unit_filter_query(super().get_query())

    def get_count_query(self) -> Query:
        return self.unit_filter_query(super().get_count_query())

    def mutate_form(self, model: UM | None, form: Form, is_created: bool) -> None:
        if not config.BUNITS or units.current_unit():
            try:
                del form['unit_id']
            except KeyError:  # pragma: no cover
                pass

    def mutate_form_opts(self, form_opts: FormOpts, is_created: bool) -> None:
        if not config.BUNITS or units.current_unit():
            if form_opts.form_rules:
                rules = [it for it in form_opts.form_rules.rules if it.field_name != 'unit_id']
                form_opts.form_rules = RuleSet(form_opts.form_rules.view, rules)

    def on_model_change(self, form: Form, model: UM, is_created: bool) -> None:
        unit_id = units.current_unit_id()
        if is_created and unit_id:
            model.unit_id = unit_id
        super().on_model_change(form, model, is_created)


password_length_v = Length(min=config.AUTH_PASSWORD_MIN_LENGTH, max=100)


class ReplicatedMixin(Generic[RM]):
    def delete_model(self, model: RM) -> bool:
        try:
            replicated.delete_model(model, user=current_user.username)
            return True
        except Exception as ex:  # pragma: no cover
            if not cast(ModelView[RM], self).handle_view_exception(ex):
                flash(gettext('Failed to delete record. %(error)s', error=str(ex)), 'error')
                logger.exception('Failed to delete record.')
            return False

    def after_model_change(self, form: Form, model: RM, is_created: bool) -> None:
        replicated.sync_model(model, current_user.username)


class UserView(ReplicatedMixin[User], BaseView[User]):
    permissions = make_permissions(perms.VIEW, perms.MANAGE_USERS)

    column_list = ('username', 'description', 'readonly')
    column_labels = dict(username='Username', description='Description', readonly='Readonly')

    column_exclude_list = ['password_hash']

    column_sortable_list = ('username', 'description', 'readonly')
    column_searchable_list = ['username']

    form_columns = ('username', 'password', 'readonly', 'description')
    form_extra_fields = dict(password=PasswordField(label='Password', validators=[OptionalV(), password_length_v]))

    def on_model_change(self, form: Form, model: User, is_created: bool) -> None:
        if model.password:  # pragma: no branch
            model.set_password(model.password)
        super().on_model_change(form, model, is_created)

    def mutate_form(self, model: User | None, form: Form, is_created: bool) -> None:
        if is_created:
            form.password.validators = [DataRequired(), password_length_v]  # type: ignore[attr-defined]


def feed_names() -> list[str]:
    return Feed.get_all_names(units.current_unit_id())


class KeyView(ReplicatedMixin[KcKey], UnitBaseView[KcKey]):
    permissions = make_permissions(perms.VIEW, perms.MANAGE_KEYS)

    def on_model_change(self, form: Form, model: KcKey, is_created: bool) -> None:
        # allow user to leave key field empty and fill it with autogenerated text
        if not model.key:
            model.key = get_unique_id()

        model.set_products(form.products.raw_data)  # type: ignore[attr-defined]
        model.validate()
        super().on_model_change(form, model, is_created)

    def on_form_prefill(self, form: Form, id: Any) -> None:
        form.products.data = json.loads(form.products.data or '[]')  # type: ignore[attr-defined] # stored as str in DB

    def handle_view_exception(self, exc: Exception) -> bool:
        if isinstance(exc, db.session.IntegrityError):  # type: ignore[attr-defined]
            error_msg = str(exc)

            if 'unique' in error_msg.lower():  # pragma: no branch
                error_msg = "Key already exists"

            flash(error_msg, category='error')
            return True

        return super().handle_view_exception(exc)  # pragma: no cover TODO

    column_list = ('key', 'note', 'server_limit', 'feed', 'products')
    column_labels = dict(
        key='Key',
        feed='Feed',
        note='Description',
        server_limit='Server Count / Limit',
        products='Products',
    )

    column_sortable_list = ('key', 'note', 'feed')
    column_searchable_list = ['key', 'feed', 'note']

    column_formatters = dict(
        key=key_name_formatter,
        server_limit=server_limit_formatter,
        feed=feed_formatter,
        products=products_formatter,
    )

    form_columns = ('key', 'note', 'server_limit', 'feed', 'products')

    # remove all default primary key validators
    form_extra_fields = dict(key=StringField(label='Key', description='Leave empty to get autogenerated key name'))
    form_edit_rules = ('note', 'server_limit', 'feed', 'products')  # do not allow user to change `key` primary key

    form_overrides = dict(feed=SelectField, products=Select2FieldMultiple)

    form_args = dict(
        server_limit=dict(label='Server Limit', validators=[NumberRange(min=0)]),
        feed=dict(choices=Choices(feed_names), coerce=lambda value: value if value != DEFAULT_FEED else None),
        products=dict(render_kw={'multiple': 'multiple'}, choices=list(pc.Product.titles.items())),
    )

    def _show_missing_fields_warning(self, text: str) -> None:
        pass


class FeedView(ReplicatedMixin[Feed], UnitBaseView[Feed]):
    permissions = make_permissions(perms.VIEW, perms.MANAGE_FEEDS)

    list_template = 'delete_feed.html'

    column_list = ('name', 'auto', 'deploy_after', 'channel')
    column_labels = dict(name='Name', auto='Autoupdate', deploy_after='Deploy after X hours')
    column_formatters = dict(
        channel=channel_formatter,
    )

    column_sortable_list = ('name',)
    column_searchable_list = ['name']

    form_columns = ('name', 'auto', 'deploy_after', 'channel')
    form_edit_rules = ('auto', 'deploy_after', 'channel')  # do not allow user to change `name` primary key
    form_overrides = dict(channel=SelectField)
    form_args = dict(
        name=dict(validators=[NoneOf(DEFAULT_FEED, message=f"Feed '{DEFAULT_FEED}' is reserved")]),
        channel=dict(choices=list(pc.Channel.titles.items())),
    )

    def _show_missing_fields_warning(self, text: str) -> None:
        pass


@formatter
def server_tags_formatter(value: list[KcServerTag]) -> str:
    return ', '.join(st.tag.name + (f':{st.tag_value}' if st.tag_value else '') for st in value if st.tag)


class ServerView(ReplicatedMixin[KcServer], BaseView):
    permissions = make_permissions(perms.VIEW, perms.MANAGE_SERVERS)

    can_edit = False
    can_create = False
    can_view_details = True
    can_export = True

    details_template = 'server_details.html'

    class FilterInactive(IntGreaterFilter):
        def apply(self, query: Query, value: int, alias: Any | None = None) -> Query:
            since_date = datetime.now() - timedelta(days=value)
            return query.filter(
                or_(
                    KcServer.last < since_date,
                    and_(KcServer.last.is_(None), KcServer.created < since_date),
                )
            )

    class FilterFeed(FilterEqual):
        def apply(self, query: Query, value: str | None, alias: Any | None = None) -> Query:
            if value == DEFAULT_FEED:
                value = None
            return query.filter(KcKey.key == KcServer.kc_key, KcKey.feed == value)

    class ServerTagFilter(FilterInList):
        def __init__(self, inverse: bool = False, **kwargs: Any) -> None:
            super().__init__(**kwargs)
            self._inverse = inverse

        def apply(self, query: Query, value: list[int], alias: Any | None = None) -> Query:
            st = aliased(KcServerTag)
            subq = exists().where(st.server_id == self.get_column(alias)).where(st.tag_id.in_(value))
            return query.filter(~subq if self._inverse else subq)

        def operation(self) -> Any:
            return lazy_gettext(('not ' if self._inverse else '') + 'in list')

    class ServerTagValueFilter(FilterLike):
        def apply(self, query: Query, value: str, alias: Any | None = None) -> Query:
            st = aliased(KcServerTag)
            subq = (
                exists().where(st.server_id == self.get_column(alias)).where(st.tag_value.ilike(parse_like_term(value)))
            )
            return query.filter(subq)

    # it's workaround which provides ability to get servers with userspace patches
    # todo: rework it to use combined model kc_servers + kc_ucheckins during refactoring
    class FilterPatchsetEqual(FilterEqual):
        def apply(self, query: Query, value: str, alias: Any | None = None) -> Query:
            if value.startswith('U'):
                last = datetime.now() - timedelta(days=db.models.DAY_DELTA)
                f = '''
                    kc_servers.server_id IN (
                        SELECT server_id from kc_ucheckins WHERE patchset = :patchset
                        AND last >= :last
                    )
                    '''
                result = query.filter(text(f).bindparams(patchset=value, last=last))
            else:
                result = query.filter(KcServer.patchset == value)
            return result

    list_template = 'server_statistics.html'

    column_list = (
        'remote_ip',
        'node',
        'server_tags',
        'euname',
        'created',
        'last',
        'server_id',
        'kc_key',
        'patchset',
        'kc_key_rel.unit.title',
    )

    column_details_list = column_export_list = [
        'remote_ip',
        'node',
        'euname',
        'created',
        'last',
        'server_id',
        'patch_level',
        'kernel_id',
        'distro_version',
        'distro',
        'machine',
        'processor',
        'release',
        'version',
        'kcare_version',
        'virt',
        'uptime',
        'ltimestamp',
        'patch_type',
        'patchset',
        'kc_key',
        'kc_key_rel.unit.title',
    ]

    column_extra_row_actions = [TemplateLinkRowAction('custom_row_actions.server_tags', gettext('Server tags'))]

    named_filter_urls = True

    column_labels = {
        'last': 'Check In',
        'node': 'Hostname',
        'server_tags': 'Tags',
        'euname': 'Effective Kernel',
        'created': 'Registered',
        'ltimestamp': 'Last Updated',
        'remote_ip': 'IP',
        'server_id': 'Server ID',
        'kc_key': 'Key',
        'kc_key_rel.unit.title': 'Business Unit',
        'name': 'Tag',
        'tag_value': 'Tag value',
    }

    column_sortable_list = (
        'remote_ip',
        'node',
        'euname',
        'created',
        'last',
        'server_id',
        'kc_key',
        'patchset',
        'kc_key_rel.unit.title',
    )

    column_default_sort = ('created', False)

    column_filters = [
        FilterEqual(column=KcServer.kc_key, name='Key'),
        FilterLike(column=KcServer.remote_ip, name='IP'),
        FilterEqual(column=KcServer.remote_ip, name='IP'),
        FilterLike(column=KcServer.node, name="Hostname"),
        FilterEqual(column=KcServer.node, name="Hostname"),
        FilterEqual(column=KcServer.server_id, name='Server ID'),
        FilterLike(column=KcServer.euname, name='Effective Kernel'),
        FilterEqual(column=KcServer.euname, name='Effective Kernel'),
        DateEqualFilter(column=KcServer.last, name='Check In'),
        DateGreaterFilter(column=KcServer.last, name='Check In'),
        DateSmallerFilter(column=KcServer.last, name='Check In'),
        DateEqualFilter(column=KcServer.created, name='Registered'),
        DateGreaterFilter(column=KcServer.created, name='Registered'),
        DateSmallerFilter(column=KcServer.created, name='Registered'),
        FilterInactive(column=KcServer.last, name='Days Inactive'),
        DateEqualFilter(column=KcServer.ltimestamp, name='Last Updated'),
        DateGreaterFilter(column=KcServer.ltimestamp, name='Last Updated'),
        DateSmallerFilter(column=KcServer.ltimestamp, name='Last Updated'),
        FilterLike(column=KcServer.patch_type, name="Patch Type"),
        FilterEqual(column=KcServer.patch_type, name="Patch Type"),
        FilterFeed(column='feed', name='Feed'),
        FilterPatchsetEqual(column='patchset', name='Patchset'),
        ServerTagFilter(column=KcServer.server_id, options=[], name='Tag'),
        ServerTagFilter(inverse=True, column=KcServer.server_id, options=[], name='Tag'),
        ServerTagValueFilter(column=KcServer.server_id, name='Tag value'),
    ]

    column_searchable_list = ['kc_key', 'server_id', 'node', 'remote_ip', KcServerTag.tag_value, Tag.name]

    column_formatters = {
        'last': local_date_formatter,
        'created': local_date_formatter,
        'ltimestamp': local_date_formatter,
        'kc_key_rel.unit.title': business_unit_formatter,
        'server_tags': server_tags_formatter,
    }

    @property
    def _template_args(self) -> dict[str, int]:
        checking_in_hours = 30 * 24
        in_past_time = datetime.now() - timedelta(hours=checking_in_hours)

        unit_id = units.current_unit_id()

        query = db.session.query(
            func.count(KcServer.server_id),
            func.count(case((KcServer.last >= in_past_time, KcServer.server_id), else_=None)),
        )

        if unit_id:
            query = query.join(KcKey, KcServer.kc_key == KcKey.key).filter(KcKey.unit_id == unit_id)

        total, checked_in = query.one()
        silent = total - checked_in

        return {'total': total, 'checked_in': checked_in, 'silent': silent}

    def list_columns_ex(self, columns: list[tuple[str, str]]) -> list[tuple[str, str]]:
        return UnitBaseView.discard_bunit_column_when_required(columns)

    def export_columns_ex(self, columns: list[tuple[str, str]]) -> list[tuple[str, str]]:
        tag_columns = [(f'tag_{tag.tag_id}', f'Tag {tag.name}') for tag in get_tags(unit_id=units.current_unit_id())[1]]
        return UnitBaseView.discard_bunit_column_when_required(columns) + tag_columns

    def details_columns_ex(self, model: KcServer, columns: list[tuple[str, str]]) -> list[tuple[str, str]]:
        tag_columns = [(f'tag_{st.tag_id}', f'Tag {st.tag.name}') for st in model.server_tags if st.tag]
        return UnitBaseView.discard_bunit_column_when_required(columns) + tag_columns

    def _get_tag_column_value(self, model: KcServer, column_suffix: str) -> str | Literal[True]:
        try:
            tag_id = int(column_suffix)
            return next(st.tag_value or True for st in model.server_tags if st.tag_id == tag_id)
        except StopIteration:
            return ''

    def get_export_value(self, model: KcServer, name: str) -> Any:
        if name.startswith('tag_'):
            return self._get_tag_column_value(model, name[4:])
        else:
            return super().get_export_value(model, name)

    @pass_context
    def get_detail_value(self, context: Context, model: KcServer, name: str) -> Any:
        if name.startswith('tag_'):
            return self._get_tag_column_value(model, name[4:])
        else:
            return super().get_detail_value(context, model, name)

    def _get_filter_groups(self) -> OrderedDict:
        '''Update Tag filter options on every request'''

        filters = super(ServerView, self)._get_filter_groups()
        for tag_filter in filters['Tag']:
            tag_filter['options'] = get_tag_filter_options(units.current_unit_id())
        return filters

    def unit_filter_query(self, q: Query, is_count: bool) -> Query:
        unit_id = units.current_unit_id()
        if config.BUNITS:
            q = q.join(KcKey, isouter=True)

        if not is_count and config.BUNITS and not unit_id:
            q = q.options(joinedload(KcServer.kc_key_rel))

        if unit_id:
            q = q.filter(KcKey.unit_id == unit_id)

        return q

    def get_query(self) -> Query:
        return self.unit_filter_query(super().get_query(), False).group_by(KcServer.server_id)

    def get_count_query(self) -> Query:
        # joining with tags causes records multiplication, so we need grouping by server_id when counting
        count_query = db.session.query(func.count(distinct(KcServer.server_id))).select_from(KcServer)
        return self.unit_filter_query(count_query, True)

    def _get_search_query(self, query: Query, args: ViewArgs) -> Query:
        '''Build query from search params as in get_list()'''
        joins: dict = {}
        if self._search_supported and args.search:
            query, _, joins, _ = self._apply_search(query, None, joins, None, args.search)
        if args.filters and self._filters:
            query, _, joins, _ = self._apply_filters(query, None, joins, None, args.filters)
        for j in self._auto_joins:
            query = query.options(joinedload(j))  # pragma: no cover Not currently used in the model
        return query

    @action('delete', lazy_gettext('Delete'), lazy_gettext('Are you sure you want to delete selected records?'))  # type: ignore[misc]
    def action_delete(self, ids: list[str]) -> None:
        '''Overrides group delete action handler to support all-pages selection and fast_mass_delete'''
        try:
            if ids:
                count = replicated.delete_servers(
                    server_ids=ids, unit_id=units.current_unit_id() or None, user=current_user.username
                )
            else:
                view_args = self._get_list_extra_args()
                search_query = self._get_search_query(self.get_query(), view_args)
                server_id_subquery = db.session.query(text('server_id')).select_from(search_query.subquery())
                ids = [row[0] for row in server_id_subquery.all()]
                count = sum(
                    replicated.delete_servers(
                        server_ids=chunk,
                        unit_id=units.current_unit_id() or None,
                        user=current_user.username,
                        audit_params={'search': view_args.search, 'filters': view_args.filters},
                    )
                    for chunk in iter_chunks(ids, size=10000)
                )

            flash(
                ngettext(
                    'Record was successfully deleted.',
                    '%(count)s records were successfully deleted.',
                    count,
                    count=count,
                ),
                'success',
            )
        except Exception as ex:
            logger.exception('Failed to delete servers')
            flash(gettext('Failed to delete records. %(error)s', error=str(ex)), 'error')

    def _get_tags_available_for_server(self, model: KcServer) -> dict[int, Tag]:
        server_unit_id = model.kc_key_rel.unit_id if model.kc_key_rel else 0
        unit_id = units.current_unit_id() or server_unit_id
        _, tags = get_tags(unit_id)

        return {tag.tag_id: tag for tag in tags}

    def _handle_server_tags_view_post(self, model: KcServer, return_url: str) -> Response:
        form = ServerTagsForm(formdata=request.form)
        if not self.validate_form(form):
            return self.render(
                'server_tags.html',
                form=form,
                all_tags=self._get_tags_available_for_server(model),
                return_url=return_url,
            )

        try:
            replicated.tag_server(
                server_id=model.server_id,
                unit_id=units.current_unit_id(),
                user=current_user.username,
                tag_list=[(t.tag_id.data, t.tag_value.data or None) for t in form.tags if t.assigned.data],
            )
            flash(gettext('Tag assignments were successfully saved.'), 'success')
        except Exception as ex:
            logger.exception(f'Failed to update tags for server {model.server_id}')
            if not self.handle_view_exception(ex):  # pragma: no branch
                flash(gettext('Failed to update tags. %(error)s', error=str(ex)), 'error')

        return cast(Response, redirect(return_url))

    def _handle_server_tags_view_get(self, model: KcServer, return_url: str) -> Response:
        all_tags = self._get_tags_available_for_server(model)
        assigned_tags = {st.tag_id: st for st in model.server_tags}
        form_data = [
            {
                'tag_id': tag_id,
                'assigned': tag_id in assigned_tags,
                'tag_value': assigned_tags[tag_id].tag_value if tag_id in assigned_tags else None,
            }
            for tag_id in all_tags.keys()
        ]
        form = ServerTagsForm(data=form_data)
        return self.render('server_tags.html', form=form, all_tags=all_tags, return_url=return_url)

    @expose('/tags/', methods=('GET', 'POST'))
    def server_tags_view(self) -> Response:
        return_url = get_redirect_target() or self.get_url('.index_view')
        server_id = get_mdict_item_or_list(request.args, 'id')
        if server_id is None:
            return cast(Response, redirect(return_url))

        model = self.get_one(server_id)
        if model is None:
            flash(gettext('Record does not exist.'), 'error')
            return cast(Response, redirect(return_url))

        if request.method == 'POST':
            return self._handle_server_tags_view_post(model, return_url)
        else:
            return self._handle_server_tags_view_get(model, return_url)


class IndexView(AdminIndexView):
    @expose('/')
    @expose('/<path>')
    def index(self, path: Any | None = None) -> Response:
        return cast(Response, redirect(url_for('kcserver.index_view')))


class AuditView(UnitBaseView[AuditLog]):
    permissions = make_permissions(perms.AUDIT)

    can_create = False
    can_delete = False
    can_edit = False
    can_view_details = False

    column_default_sort = ('date', True)
    column_searchable_list = ('details',)
    column_display_actions = False

    named_filter_urls = True

    column_list = ('date', 'user', 'action', 'details')
    column_sortable_list = ('date', 'user', 'action')

    column_filters = (
        FilterEqual(column=audit.AuditLog.user, name='User'),
        FilterEqual(column=audit.AuditLog.action, name='Action'),
    )

    column_formatters = dict(
        date=local_date_formatter,
        user=filter_formatter,
        action=filter_formatter,
    )


def unit_url_value(slug: str, url: str) -> str:
    return f'/{escape(slug)}{url}'


class UnitView(ReplicatedMixin[Unit], BaseView[Unit]):
    permissions = make_permissions(
        {perms.MANAGE_UNIT, perms.VIEW_UNIT}, perms.MANAGE_UNIT, edit_view={perms.MANAGE_UNIT, perms.EDIT_UNIT_TITLE}
    )

    column_list = ('slug', 'login_url', 'api_url', 'title')
    column_labels = {'slug': 'Name'}
    column_formatters = {
        'login_url': mformatter(lambda m: Markup(f'<a href="{unit_url_value(m.slug, "/admin/login")}">login url</a>')),
        'api_url': mformatter(lambda m: Markup(f'<a href="{unit_url_value(m.slug, "/admin/api")}">API url</a>')),
    }

    form_args = {
        'slug': {
            'description': 'A unique ID used to identify this Business Unit',
            'validators': [NoneOf(['_'], message="Name '_' is reserved")],
        },
    }

    def mutate_form(self, model: Unit | None, form: Form, is_created: bool) -> None:
        if not perms.user_has_permission(perms.MANAGE_UNIT):
            del form['slug']

    def get_one(self, id: str) -> Unit | None:
        model: Unit | None = super().get_one(id)
        if not perms.user_has_permission(perms.MANAGE_UNIT):
            if model and model.id != units.current_unit_id():
                return None
        return model

    def unit_filter_query(self, q: Query) -> Query:
        if unit_id := units.current_unit_id():
            q = q.filter_by(id=unit_id)
        return q

    def get_query(self) -> Query:
        return self.unit_filter_query(super().get_query())

    def get_count_query(self) -> Query:
        return self.unit_filter_query(super().get_count_query())


class UnitUserView(ReplicatedMixin[UnitUser], UnitBaseView[UnitUser]):
    permissions = make_permissions(perms.MANAGE_UNIT_USERS, perms.MANAGE_UNIT_USERS)

    unit_exclude_default = True

    column_exclude_list = ['password_hash']
    column_searchable_list = ['username']
    column_list = ['username', 'role', 'description']
    column_formatters = dict(role=formatter(lambda value: perms.UnitRoles.get(value).title))

    form_columns = ('username', 'password', 'role', 'description')
    form_extra_fields = dict(password=PasswordField(label='Password', validators=[OptionalV(), password_length_v]))

    form_overrides = dict(role=SelectField)

    form_args = dict(role=dict(choices=perms.UnitRoles.titles.items()))

    def on_model_change(self, form: Form, model: UnitUser, is_created: bool) -> None:
        if model.password:  # pragma: no branch
            model.set_password(model.password)
        super().on_model_change(form, model, is_created)

    def mutate_form(self, model: UnitUser | None, form: Form, is_created: bool) -> None:
        if is_created:
            form.password.validators = [DataRequired(), password_length_v]  # type: ignore[attr-defined]
        super().mutate_form(model, form, is_created)


def _is_super_admin() -> bool:
    return not units.current_unit() and perms.SUPERADMIN in current_user.permissions


@formatter
def expiration_date_formatter(value: datetime | None) -> str:
    if not value:
        return ''
    local_date = convert_utc_date_to_local(value).strftime(config.DATETIME_FORMAT)
    if value <= datetime.now(timezone.utc).replace(tzinfo=None):
        return Markup(f'<span class="text-danger">{local_date}</span>')
    return local_date


class ApiKeyView(ReplicatedMixin[ApiKey], BaseView[ApiKey]):
    permissions = make_permissions(perms.VIEW, perms.VIEW)
    can_edit = False
    column_list = ('name', 'username', 'created', 'expiration', 'unit_user.unit.title')
    column_labels = {
        'name': 'Key Name',
        'username': 'Username',
        'created': 'Created date',
        'expiration': 'Expiration date',
        'unit_user.unit.title': 'Business Unit',
    }
    column_sortable_list = column_list
    column_searchable_list = ('name', 'username')
    form_columns = ('name', 'expiration')
    form_args = {
        'expiration': {
            'description': 'Leave this field blank for a key that never expires',
            'default': lambda: datetime.now() + timedelta(days=90),
        }
    }

    column_formatters = {
        'unit_user.unit.title': business_unit_formatter,
        'created': local_date_formatter,
        'expiration': expiration_date_formatter,
    }

    column_default_sort = ('created', True)

    def list_columns_ex(self, columns: list[tuple[str, str]]) -> list[tuple[str, str]]:
        if config.BUNITS and _is_super_admin():
            return columns
        return columns[:-1]

    def apply_visibility_filter(self, query: Query) -> Query:
        if unit_id := units.current_unit_id():
            if perms.user_has_permission(perms.MANAGE_UNIT_USERS):
                # Unit admin
                return query.join(ApiKey.unit_user).filter_by(unit_id=unit_id)
            # Unit user
            return query.filter(ApiKey.unit_user_id == current_user.id)
        if _is_super_admin():
            return query.filter(ApiKey.unit_user_id.is_(None)) if not config.BUNITS else query
        # Viewer
        return query.filter(ApiKey.kc_user_id == current_user.id)

    def get_query(self) -> Query:
        return self.apply_visibility_filter(super().get_query())

    def get_count_query(self) -> Query:
        return self.apply_visibility_filter(super().get_count_query())

    def get_one(self, id: str) -> ApiKey | None:
        return self.get_query().filter(ApiKey.id == id).first()

    def mutate_form(self, model: ApiKey | None, form: Form, is_created: bool) -> None:
        form.name.validators = [DataRequired()]  # type: ignore[attr-defined]
        return super().mutate_form(model, form, is_created)

    def on_model_change(self, form: Form, model: ApiKey, is_created: bool) -> None:
        if is_created:
            token = ApiKey.generate()
            if isinstance(current_user, UnitUser):
                model.unit_user_id = current_user.id
            elif isinstance(current_user, User):
                model.kc_user_id = current_user.id
            else:  # pragma: no cover
                raise ValueError(f'{type(current_user)} is not supported')
            model.key_hash = ApiKey.get_token_hash(token)
            g.api_key = token
        else:  # pragma: no cover
            raise NotImplementedError('API Keys are not editable')

        if model.expiration:  # pragma: no branch
            model.expiration = convert_local_date_to_utc(model.expiration)

        return super().on_model_change(form, model, is_created)

    def after_model_change(self, form: Form, model: ApiKey, is_created: bool) -> None:
        if hasattr(g, 'api_key'):  # pragma: no branch
            flash(
                'Please save the token below. For security reasons, you will not be able to view it again.',
                category='info',
            )
            flash(g.api_key, category='info')
        return super().after_model_change(form, model, is_created)


class TagView(ReplicatedMixin[Tag], UnitBaseView[Tag]):
    permissions = make_permissions(perms.VIEW, perms.MANAGE_SERVERS)
    column_list = ('name', 'server_count')
    column_labels = {'name': 'Tag Name', 'server_count': 'Number of servers'}
    column_sortable_list = column_list
    column_searchable_list = ('name',)
    column_default_sort = ('name', False)

    form_columns = ('name',)
    form_args = {
        'name': {'validators': (Regexp(r'^[\w+-]+$', message="Only alphanumeric characters and [_+-] are valid here"),)}
    }


admin = Admin(app, index_view=IndexView(), template_mode='bootstrap3', base_template='base.html')

admin.add_view(UserView(User, db.session, 'Users'))
admin.add_view(KeyView(KcKey, db.session, 'Keys'))
admin.add_view(FeedView(Feed, db.session, 'Feeds'))
admin.add_view(ServerView(KcServer, db.session, 'Servers'))
admin.add_view(AuditView(audit.AuditLog, db.session, 'User Log'))
admin.add_view(UnitView(Unit, db.session, 'Units'))
admin.add_view(UnitUserView(UnitUser, db.session, 'Unit users'))
admin.add_view(ApiKeyView(ApiKey, db.session, 'API keys'))
admin.add_view(TagView(Tag, db.session, 'Server tags'))
