from __future__ import annotations

from itertools import groupby
from operator import itemgetter
from typing import Any, Iterable, cast

import config
from flask import Request, Response, request
from flask_admin import BaseView, expose
from flask_admin.babel import ngettext
from flask_login import login_required

from eportal.analytics import ServersStats, BaseFilter, BUnitFilter, TagInFilter, TagValueLikeFilter, DaysInactiveFilter
from eportal.db import session, models
from eportal.web import units

from .admin import admin, get_tag_filter_options


class Filters:
    '''filters implementation compatible with flask-admin model view'''

    def __init__(self) -> None:
        self._filters: dict[str, dict[str, Any]] = {}

    def add_filter(self, flt: BaseFilter, options: Any = None, type: str | None = None) -> None:
        arg_name = f'{flt.name}_{flt.operation.value}'.lower().replace(' ', '_')
        self._filters[arg_name] = {
            'arg': arg_name,
            'flt': flt,
            'index': len(self._filters),
            'options': options,
            'type': type,
            'name': flt.name,
            'operation': flt.operation.value,
        }

    def get_filters(self) -> dict[str, list[dict[str, Any]]]:
        def make_filters_group(data: Iterable[dict[str, Any]]) -> list[dict[str, Any]]:
            return [{k: v for k, v in d.items() if k in ('operation', 'arg', 'index', 'options', 'type')} for d in data]

        filters = sorted(self._filters.values(), key=lambda x: (x['name'], x['index']))
        return {name: make_filters_group(flt_group) for name, flt_group in groupby(filters, key=itemgetter('name'))}

    def get_active_filters(self, request: Request) -> list[tuple[BaseFilter, str, int, int]]:
        '''Returns [(filter, value, index, pos)]'''

        def parse_filter_arg(arg: str) -> tuple[BaseFilter, str, int, int] | None:
            pos, *arg_name = arg[3:].split('_', 1)
            if not arg_name or (flt := self._filters.get(arg_name[0])) is None:
                return None
            return (
                cast(BaseFilter, flt['flt']),
                request.args[arg],
                cast(int, flt['index']),
                int(pos) if pos.isdigit() else 0,
            )

        return list(filter(None, (parse_filter_arg(arg) for arg in request.args if arg.startswith('flt'))))


class HealthMonitorView(BaseView):
    _template = 'health_monitor.html'

    def __init__(self) -> None:
        super().__init__(name='Dashboard', endpoint='health_monitor', url='health')  # type: ignore[call-arg]

    @expose('/')
    @login_required
    def index(self) -> Response:
        unit_id = units.current_unit_id()
        filters = self._create_filters(unit_id)
        active_flt_data = filters.get_active_filters(request)
        return self.render(
            self._template,
            filter_groups=filters.get_filters(),
            active_filters=[(f[2], f[0].name, f[1]) for f in sorted(active_flt_data, key=itemgetter(3))],
            extra_args={},
            sort_column=None,
            clear_search_url=self.get_url('.index'),  # type: ignore[attr-defined]
            **self._dashboard_data([(flt, val) for flt, val, _, _ in active_flt_data], unit_id),
        )

    def _create_filters(self, unit_id: int) -> Filters:
        filters = Filters()
        if config.BUNITS and unit_id == 0:
            bu_options = units.titles()
            filters.add_filter(BUnitFilter(BUnitFilter.Operation.IN), options=bu_options, type='select2-tags')
            filters.add_filter(BUnitFilter(BUnitFilter.Operation.NOT_IN), options=bu_options, type='select2-tags')
        tags_options = get_tag_filter_options(unit_id)
        filters.add_filter(TagInFilter(), options=tags_options, type='select2-tags')
        filters.add_filter(TagInFilter(inverse=True), options=tags_options, type='select2-tags')
        filters.add_filter(TagValueLikeFilter())
        filters.add_filter(DaysInactiveFilter(DaysInactiveFilter.Operation.SMALLER))
        filters.add_filter(DaysInactiveFilter(DaysInactiveFilter.Operation.GREATER))
        return filters

    def _dashboard_data(self, active_filters: list[tuple[BaseFilter, str]], unit_id: int) -> dict[str, Any]:
        if config.BUNITS and unit_id:
            active_filters.append((BUnitFilter(BUnitFilter.Operation.EQUALS), str(unit_id)))
        stats = ServersStats(active_filters)

        total_servers = session.query(models.KcServer).count()
        selected_servers, uptime_data = stats.get_uptime_stats()
        distro_stats = self._truncate_results(stats.get_distro_stats())
        arch_stats = stats.get_arch_stats()

        return {
            'total_servers': total_servers,
            'selected_servers': selected_servers,
            'uptime': {
                k: self._human_readable_period(v) for k, v in uptime_data.items() if k in ('average', 'max', 'median')
            },
            'os_distribution': distro_stats,
            'uptime_distribution_data': {
                'labels': ['under 6 months', '6 to 12 months', '12 to 18 months', 'over 18 months'],
                'data': [uptime_data[k] for k in ('m0', 'm6', 'm12', 'm18')],
            },
            'arch_distribution_data': {
                'labels': [name for name, _ in arch_stats],
                'data': [count for _, count in arch_stats],
            },
        }

    def _truncate_results(
        self, data: list[tuple[str, int, int]], limit: int = 10, others_title: str = 'others'
    ) -> list[tuple[str, int, int]]:
        if len(data) <= limit:
            return data
        active_count = silent_count = 0
        for _, active, silent in data[limit - 1 :]:
            active_count += active
            silent_count += silent
        return data[0 : limit - 1] + [(others_title, active_count, silent_count)]

    def _human_readable_period(self, seconds: int | None) -> str:
        if seconds is None:
            return 'N/A'
        minutes = (seconds + 30) // 60
        if minutes < 60:
            return str(ngettext('1 minute', '%(minutes)s minutes', minutes, minutes=minutes))
        hours = (minutes + 30) // 60
        if hours < 24:
            return str(ngettext('1 hour', '%(hours)s hours', hours, hours=hours))
        days = (hours + 12) // 24
        if days < 30:
            return str(ngettext('1 day', '%(days)s days', days, days=days))
        months = (days + 15) // 30
        return str(ngettext('1 month', '%(months)s months', months, months=months))


admin.add_view(HealthMonitorView())
